#pragma once

#include <string>
#include <vector>
#include <unordered_map>
#include <queue>

class Gate {
public:
    int topo;
    std::string name;
    enum Type { AND, NAND, OR, NOR, XOR, XNOR, NOT, BUF, INPUT } type;
    int value;
    bool sa[2];
    bool stem;
    bool isPI;
    bool isPO;

    std::unordered_map<Gate*, std::pair<int, int>> sa_by_out;

    std::vector<Gate*> pre_stems;

    std::vector<Gate*> block;

    std::vector<Gate*> outputs;
    std::vector<Gate*> inputs;

    bool is_propagated();
    int cal_value();
};

class Fault {
public:
    Gate* gate;
    enum Type { SA0, SA1 } type;
    bool detected;
    Fault(Gate* gate, Type type):gate(gate),type(type),detected(0) {}
};

class Circuit {
public:
std::vector<Gate*> PIs;
std::vector<Gate*> POs;
std::vector<Gate*> gates;
std::vector<Gate*> stems; // PI + stems

std::unordered_map<std::string, Gate*> name2gate;

std::queue<Gate*> tmp;
std::unordered_map<Gate*, bool> tmp_used;

void parse_from_file(const char *filename);
void print_gates();

bool is_valid_circuit();

void init_topo_index();
void init_stems();


// local search

std::vector<int> local_search(const std::vector<Fault*> &faults);

void ls_init_circuit();
void ls_init_vector(std::vector<Gate*> *vec);
void ls_flip(Gate* stem);

};